/*
**  ConsoleWindowController.m
**
**  Copyright (c) 2001-2004
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ConsoleWindowController.h"

#include "Constants.h"
#include "GNUMail.h"
#include "MailboxManagerController.h"
#include "Task.h"
#include "TaskManager.h"

#ifndef MACOSX
#include "ConsoleWindow.h"
#endif

#include <Pantomime/CWMessage.h>
#include <Pantomime/CWURLName.h>

//
// NOTE: For a good descriptions on what are the available "tasks" and how
//       each one of them is used, please read the documentation on top
//       of the TaskManager.m file.
//

static ConsoleWindowController *singleInstance = nil;

//
//
//
@interface ProgressIndicatorCell : NSTextFieldCell
{
  Task *_task;
}
- (void) setTask: (Task *) theTask;
@end


@implementation ProgressIndicatorCell

- (id) copyWithZone: (NSZone *) theZone
{
  return [[ProgressIndicatorCell alloc] init];
}

- (void) drawWithFrame: (NSRect) cellFrame 
		inView: (NSView *) controlView
{
  float f, w, h, x, y;
  int op;
  
  if (_task == nil)
    {
      return;
    }

  op = [_task op];
  
  if (op == RECEIVE_IMAP)
    {
      f = (float)_task->received_count/(float)_task->total_count;
    }
  else 
    {
      f = _task->current_size/_task->total_size;
    }

  w = cellFrame.size.width;
  h = cellFrame.size.height;
  x = cellFrame.origin.x;
  y = cellFrame.origin.y;
  
  cellFrame.origin.y += 3;
  cellFrame.size.height = 20;
  NSDrawDarkBezel(cellFrame, cellFrame);
  
  if ((op == RECEIVE_IMAP && f > 0) ||
      (op == SEND_SMTP && f > 0) ||
      (op == LOAD_ASYNC && f > 0) ||
      (op == SAVE_ASYNC && f > 0) ||
      (op == RECEIVE_POP3 && _task->total_count > 0))
    {
      cellFrame.size.height = 16;
      cellFrame.origin.y += 2;
      cellFrame.origin.x += 1.5;
      cellFrame.size.width = cellFrame.size.width*f - 4;

      if (cellFrame.size.width > (w-4)) cellFrame.size.width = (w-4);

      [[NSColor grayColor] set];     
      NSRectFill(cellFrame);
    }
     
  cellFrame.size.width = w-4;
  cellFrame.size.height = h;
  cellFrame.origin.x = x+4;
  cellFrame.origin.y = y;

  [super drawInteriorWithFrame: cellFrame  inView: controlView];
}


- (void) setTask: (Task *) theTask
{
  _task = theTask;
}

@end


//
//
//
@interface ConsoleWindowController (Private)
- (void) _startTask;
- (void) _stopTask;
@end

//
// 
//
@interface ConsoleMessage : NSObject
{
  @public
    NSString *message;
    NSCalendarDate *date;
}

- (id) initWithMessage: (NSString *) theMessage;

@end


//
//
//
@implementation ConsoleWindowController

- (id) initWithWindowNibName: (NSString *) windowNibName
{
  NSButtonCell *aCell;

#ifdef MACOSX
  
  self = [super initWithWindowNibName: windowNibName];
  
#else
  ConsoleWindow *aConsoleWindow;

  aConsoleWindow = [[ConsoleWindow alloc] initWithContentRect: NSMakeRect(100,100,500,280)
					  styleMask: NSTitledWindowMask|NSClosableWindowMask|NSResizableWindowMask
					  backing: NSBackingStoreBuffered
					  defer: NO];
  
  self = [super initWithWindow: aConsoleWindow];
  
  [aConsoleWindow layoutWindow];
  [aConsoleWindow setDelegate: self];

  // We link our outlets
  tasksTableView = aConsoleWindow->tasksTableView;
  messagesTableView = aConsoleWindow->messagesTableView;
  toggle = aConsoleWindow->toggle;
  delete = aConsoleWindow->delete;
  suspend = aConsoleWindow->suspend;
  save = aConsoleWindow->save;
  
  RELEASE(aConsoleWindow);
#endif

  [[self window] setTitle: _(@"GNUMail Console")];
  
  // We finally set our autosave window frame name and restore the one from the user's defaults.
  [[self window] setFrameAutosaveName: @"ConsoleWindow"];
  [[self window] setFrameUsingName: @"ConsoleWindow"];

  // We set the custom cell for the Status column
  [[tasksTableView tableColumnWithIdentifier: @"Status"] setDataCell: AUTORELEASE([[ProgressIndicatorCell alloc] init])];
  [tasksTableView setRowHeight: 26];
  [tasksTableView setIntercellSpacing: NSZeroSize];

  // We finally set the custom cell for the Operation column
  aCell = [[NSButtonCell alloc] init];
  [aCell setTitle: _(@"Stop")];
  [aCell setButtonType: NSMomentaryPushInButton];
  [[tasksTableView tableColumnWithIdentifier: @"Operation"] setDataCell: aCell];
  RELEASE(aCell);

  [self updateToggle];

  return self;
}


//
//
//
- (void) dealloc
{
#ifdef MACOSX
  [tasksTableView setDataSource: nil];
  [messagesTableView setDataSource: nil];
#endif

  RELEASE(allMessages);

  [super dealloc];
}


//
// action methods
//
- (IBAction) deleteClicked: (id) sender
{
  int aRow;
     
  aRow = [tasksTableView selectedRow];

  if (aRow >= 0)
    {
      Task *aTask;
      
      // No need to call reloadData here since it's called in -removeTask.
      aTask = [[[TaskManager singleInstance] allTasks] objectAtIndex: aRow];
      
      if ([aTask running])
	{
	  NSRunInformationalAlertPanel(_(@"Delete error!"),
				       _(@"You can't delete a running task. Stop it first."),
				       _(@"OK"),
				       NULL, 
				       NULL,
				       NULL);
	  return;
	}

      [[TaskManager singleInstance] removeTask: aTask];
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) saveClicked: (id) sender
{
  int aRow;
     
  aRow = [tasksTableView selectedRow];

  if (aRow >= 0)
    {
      CWURLName *theURLName;
      NSData *aData;
      Task *aTask;
      
      aTask = [[[TaskManager singleInstance] allTasks] objectAtIndex: aRow];

      if ([aTask running])
	{
	  NSRunInformationalAlertPanel(_(@"Save error!"),
				       _(@"You can't save the message in Drafts if the task is running. Stop it first."),
				       _(@"OK"),
				       NULL, 
				       NULL,
				       NULL);
	  return;
	}

      // We finally get our CWURLName object.
      theURLName = [[CWURLName alloc] initWithString: [[[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] objectForKey: [aTask key]] 
							 objectForKey: @"MAILBOXES"] objectForKey: @"DRAFTSFOLDERNAME"]
				      path: [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"]];
      
      if ([[aTask message] respondsToSelector: @selector(isEqualToData:)])
	{
	  aData = [aTask message];
	}
      else
	{
	  aData = [[aTask message] dataValue];
	}

      [[MailboxManagerController singleInstance] addMessage: aData
						 toFolder: theURLName];
      RELEASE(theURLName);
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) suspendClicked: (id) sender
{
  int aRow;
     
  aRow = [tasksTableView selectedRow];

  if (aRow >= 0)
    {
      Task *aTask;
      
      aTask = [[[TaskManager singleInstance] allTasks] objectAtIndex: aRow];

      if ([aTask running])
	{
	  [self _stopTask];
	}

      [aTask setDate: [NSDate distantFuture]];
      [tasksTableView setNeedsDisplay: YES];
    }
  else
    {
      NSBeep();
    }
}


//
// delegate methods
//
- (void) windowWillClose: (NSNotification *) theNotification
{
  // Do nothing
}


//
//
//
- (void) windowDidLoad
{
  allMessages = [[NSMutableArray alloc] init];

#ifdef MACOSX
  [tasksTableView reloadData];
  [messagesTableView reloadData];
#endif
}


//
// Data Source methods
//
- (int) numberOfRowsInTableView: (NSTableView *)aTableView
{
  if (aTableView == tasksTableView)
    {
      return [[[TaskManager singleInstance] allTasks] count];
    }
  else
    {
      return [allMessages count];
    }
}


//
//
//
- (id)           tableView: (NSTableView *) aTableView
 objectValueForTableColumn: (NSTableColumn *) aTableColumn 
		       row: (int) rowIndex
{
  if (aTableView == tasksTableView)
    {
      NSString *aString;
      Task *aTask;
      
      aTask = [[[TaskManager singleInstance] allTasks] objectAtIndex: rowIndex];
      
      if ([[[aTableColumn headerCell] stringValue] isEqual: _(@"Activity")])
	{
	  switch ([aTask op])
	    {
	    case RECEIVE_IMAP:
	      aString = [NSString stringWithFormat: _(@"IMAP - %@"), [aTask key]];
	      break;
	      
	    case RECEIVE_POP3:
	      aString = [NSString stringWithFormat: _(@"POP3 - %@"), [aTask key]];
	      break;

	    case SEND_SENDMAIL:
	      aString = [NSString stringWithFormat: _(@"Mailer - %@"), ([aTask sendingKey] ? [aTask sendingKey] : [aTask key])];
	      break;

	    case SEND_SMTP:
	      aString = [NSString stringWithFormat: _(@"SMTP - %@"), ([aTask sendingKey] ? [aTask sendingKey] : [aTask key])];
	      break;

	    case SAVE_ASYNC:
	      aString = [NSString stringWithFormat: _(@"Saving message...")];
	      break;

	    case LOAD_ASYNC:
	      aString = [NSString stringWithFormat: _(@"Loading message...")];
	      break;

	    default:
	      aString = @"";
	    }
	  
	  return aString;
	}
      else if ([[[aTableColumn headerCell] stringValue] isEqual: _(@"Status")])
	{
	  aString = @"";

	  if ([aTask running])
	    {
	      if ([aTask op] == RECEIVE_POP3 && aTask->total_count)
		{
		  aString = [NSString stringWithFormat: _(@"Received message %d of %d"), aTask->received_count, aTask->total_count];
		}
	      if ([aTask op] == RECEIVE_IMAP && aTask->total_count)
		{
		  aString = [NSString stringWithFormat: _(@"Got status for mailbox %d of %d"), aTask->received_count, aTask->total_count];
		}
	      else if ([aTask op] == SEND_SMTP || [aTask op] == SAVE_ASYNC || [aTask op] == LOAD_ASYNC)
		{
		  aString = [NSString stringWithFormat: _(@"Completed %0.1fKB of %0.1fKB."),
				      (aTask->current_size > aTask->total_size ? aTask->total_size : aTask->current_size),
				      aTask->total_size];
		}
	    }
	  else
	    {
	      aString = [NSString stringWithFormat: _(@"Suspended - Scheduled to run at %@"),
				  [[aTask date] descriptionWithCalendarFormat: @"%H:%M:%S"
						timeZone: nil
						locale: nil]];
	    }
	  return aString;
	}
      else
	{
	  if ([aTask running])
	    {
	      return @"Stop";
	    }
	  
	  return @"Start";
	}
    }
  else
    {
      ConsoleMessage *aMessage;
      
      aMessage = [allMessages objectAtIndex: rowIndex];
      
      if ( [[[aTableColumn headerCell] stringValue] isEqual: _(@"Date")] )
        {
          return [aMessage->date descriptionWithCalendarFormat: _(@"%H:%M:%S.%F")
			  timeZone: [aMessage->date timeZone]
			  locale: nil];
        }
      else
        {
          return aMessage->message;
        }
    }

  return nil;
}


//
//
//
- (void) tableView: (NSTableView *) aTableView
    setObjectValue: (id) anObject
    forTableColumn: (NSTableColumn *) aTableColumn
	       row: (int) rowIndex
{
  if ([[[aTableColumn headerCell] stringValue] isEqual: _(@"Operation")])
    {
      [self _stopTask];
    }
}


//
//
//
- (void) tableView: (NSTableView *)aTableView
   willDisplayCell: (id)aCell
    forTableColumn: (NSTableColumn *)aTableColumn
               row: (int)rowIndex
{
  if (aTableView == tasksTableView &&
      [[[aTableColumn headerCell] stringValue] isEqual: _(@"Status")])
    {
      [aCell setTextColor: [NSColor whiteColor]];
      [aCell setAlignment: NSCenterTextAlignment];
      [(ProgressIndicatorCell *)[aTableColumn dataCell] setTask: [[[TaskManager singleInstance] allTasks] objectAtIndex: rowIndex]];
    }
}


//
//
//
- (void) tableViewSelectionDidChange: (NSNotification *) aNotification
{
  [self updateToggle];
}


//
// access / mutation method
//
- (NSTableView *) tasksTableView
{
  return tasksTableView;
}


//
// Other methods
//
- (void) addConsoleMessage: (NSString *) theString
{
  ConsoleMessage *aMessage;
  
  aMessage = [[ConsoleMessage alloc] initWithMessage: theString];
  
  [allMessages insertObject: aMessage  atIndex: 0];
  RELEASE(aMessage);
  
  // We never keep more than 25 object in our array
  if ([allMessages count] > 25)
    {
      [allMessages removeLastObject];
    }
  
  [messagesTableView reloadData];
}


//
//
//
- (void) updateToggle
{
  Task *aTask;
  int aRow;
     
  aRow = [tasksTableView selectedRow];

  if (aRow >= 0 && [tasksTableView numberOfRows] > 0 &&
      (aTask = [[[TaskManager singleInstance] allTasks] objectAtIndex: aRow]) &&
      [aTask op] != LOAD_ASYNC &&
      [aTask op] != SAVE_ASYNC)
    {
      [toggle setEnabled: YES];
      [delete setEnabled: YES];
      [suspend setEnabled: YES];
      [save setEnabled: YES];
    
      if ([aTask running])
	{
	  [toggle setTitle: _(@"Stop")];
	  [toggle setAction: @selector(_stopTask)];
	}
      else
	{
	  [toggle setTitle: _(@"Start")];
	  [toggle setAction: @selector(_startTask)];
	}
    }
  else
    {
      [toggle setEnabled: NO];
      [delete setEnabled: NO];
      [suspend setEnabled: NO];
      [save setEnabled: NO];
    }

  [tasksTableView setNeedsDisplay: YES];
}


//
// class methods
//
+ (id) singleInstance
{
  if (singleInstance == nil)
    {
      singleInstance = [[ConsoleWindowController alloc] initWithWindowNibName: @"ConsoleWindow"];
    }

  return singleInstance;
}

@end


//
//
//
@implementation ConsoleWindowController (Private)

- (void) _startTask
{
  int count, row;

  count = [[[TaskManager singleInstance] allTasks] count];
  row = [tasksTableView selectedRow];

  if (row >= 0 && row < count)
    {
      Task *aTask;

      aTask = [[[TaskManager singleInstance] allTasks] objectAtIndex: row];
      [aTask setDate: [NSDate date]];
      [aTask setImmediate: YES];
      [[TaskManager singleInstance] nextTask];
      [toggle setTitle: _(@"Stop")];
      [toggle setAction: @selector(_stopTask)];
      [toggle setNeedsDisplay: YES];
      [tasksTableView setNeedsDisplay: YES];
    }
}

- (void) _stopTask
{
  int count, row;

  count = [[[TaskManager singleInstance] allTasks] count];
  row = [tasksTableView selectedRow];

  if (row >= 0 && row < count)
    {
      Task *aTask;

      aTask = [[[TaskManager singleInstance] allTasks] objectAtIndex: row];

      if (aTask->op != LOAD_ASYNC &&
	  aTask->op != SAVE_ASYNC)
	{
	  [[TaskManager singleInstance] stopTask: aTask];
	  [toggle setTitle: _(@"Start")];
	  [toggle setAction: @selector(_startTask)];
	  [toggle setNeedsDisplay: YES];
	  [tasksTableView setNeedsDisplay: YES];
	}
    }
}

@end



//
// 
//
@implementation ConsoleMessage

- (id) initWithMessage: (NSString *) theMessage
{
  self = [super init];

  message = RETAIN(theMessage);
  date = RETAIN([NSCalendarDate calendarDate]);

  return self;
}

- (void) dealloc
{
  RELEASE(message);
  RELEASE(date);
  [super dealloc];
}

@end
